/***************************************************************************
                          main.c  -  SDL Audio plugin for mupen64
                             -------------------
    begin                : Fri Oct 3 2003
    copyright            : (C) 2003 by Juha Luotio aka JttL
    email                : juha.luotio@porofarmi.net
    version              : 1.0.1.3
 ***************************************************************************/
/***************************************************************************
     This plug-in is originally based on Hactarux's "mupen audio plugin"
     and was modified by JttL. Actually there is no much original code
     left, but I think it's good to mention it anyway.
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 CHANGELOG:
 1.1:
 +Audio device is opened now with native byte ordering of the machine. Just
  for compatibility (thanks Flea).
 +Fixed possible double freeing bug (thanks Flea)
 +Optimizations in AiLenChanged
 +Fixed segmentation fault when changing rom.
 
 1.0.1.3:
 +Smarter versioning. No more betas.
 +More cleaning up done.
 +Buffer underrun and overflow messages appear now at stderr (if DEBUG is
  defined)
 +Many things are now precalculated (this should bring a small performance
  boost)
 +Buffer underrun bug fixed.
 +Segmentation fault when closing rom fixed (at least I think so)
  
 1.0 beta 2:
 +Makefile fixed to get rid of annoying warning messages
 +Cleaned up some old code
 +Default frequency set to 33600Hz (for Master Quest compatibility)
 +Better syncronization (needs some work still though)
 
 1.0 beta 1:
 +First public release
 

 ***************************************************************************/
/***************************************************************************
 TODO:
 +Config -file
 +GUI for adjusting config file settings ;)
 ***************************************************************************/

#ifdef USE_GTK
#include <gtk/gtk.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "SDL.h"
#include "SDL_audio.h"
#include "SDL_thread.h"
#include "winlnxdefs.h"
#include "Audio_#1.2.h"

#ifdef USE_RTC
#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#endif

/* Size of primary buffer in bytes. This is the buffer where audio is loaded
   after it's extracted from n64's memory. */
#define PRIMARY_BUFFER_SIZE 65536

/* If buffer load goes under LOW_BUFFER_LOAD_LEVEL then game is speeded up to
   fill the buffer. If buffer load exeeds HIGH_BUFFER_LOAD_LEVEL then some
   extra slowdown is added to prevent buffer overflow (which is not supposed
   to happen in any circumstanses if syncronization is working but because
   computer's clock is such inaccurate (10ms) that might happen. I'm planning
   to add support for Real Time Clock for greater accuracy but we will see.

   The plugin tries to keep the buffer's load always between these values.
   So if you change only PRIMARY_BUFFER_SIZE, nothing changes. You have to
   adjust these values instead. You propably want to play with
   LOW_BUFFER_LOAD_LEVEL if you get dropouts. */
#define LOW_BUFFER_LOAD_LEVEL 16384
#define HIGH_BUFFER_LOAD_LEVEL 32768

/* Size of secondary buffer. This is actually SDL's hardware buffer. This is
   amount of samples, so final bufffer size is four times this. */
#define SECONDARY_BUFFER_SIZE 512

/* This sets default frequency what is used if rom doesn't want to change it.
   Popably only game that needs this is Zelda: Ocarina Of Time Master Quest */
#define DEFAULT_FREQUENCY 33600


/*--------------- VARIABLE DEFINITIONS ----------------*/

/* Read header for type definition */
static AUDIO_INFO AudioInfo;
/* The hardware specifications we are using */
static SDL_AudioSpec *hardware_spec;
/* Pointer to the primary audio buffer */
static Uint8 *buffer = NULL;
/* Position in buffer array where next audio chunk should be placed */
static unsigned int buffer_pos = 0;
/* Audio frequency, this is usually obtained from the game, but for compatibility we set default value */
static int frequency = DEFAULT_FREQUENCY;
/* This is for syncronization, it's ticks saved just before AiLenChanged() returns. */
static Uint32 last_ticks = 0;
/* This is amount of ticks that are needed for previous audio chunk to be played */
static Uint32 expected_ticks = 0;



/* ----------- FUNCTIONS ------------- */
/* This function closes the audio device and reinitializes it with requested frequency */
void InitializeAudio(int freq);

EXPORT void CALL
AiDacrateChanged( int SystemType )
{
        int f = DEFAULT_FREQUENCY;
        switch (SystemType)
        {
        case SYSTEM_NTSC:
                f = 48681812 / (*AudioInfo.AI_DACRATE_REG + 1);
                break;
        case SYSTEM_PAL:
                f = 49656530 / (*AudioInfo.AI_DACRATE_REG + 1);
                break;
        case SYSTEM_MPAL:
                f = 48628316 / (*AudioInfo.AI_DACRATE_REG + 1);
                break;
        }
        InitializeAudio(f);
}


EXPORT void CALL
AiLenChanged( void )
{

        DWORD LenReg = *AudioInfo.AI_LEN_REG;
        Uint8 *p = (Uint8*)(AudioInfo.RDRAM + (*AudioInfo.AI_DRAM_ADDR_REG & 0xFFFFFF));

        if(buffer_pos + LenReg  < PRIMARY_BUFFER_SIZE)
        {
                register unsigned int i;
                SDL_LockAudio();
                for ( i = 0 ; i < LenReg ; i += 4 )
                {

                        /* These where for testing purpose only. Which made clear it's not wise to
                           modify the sound on the fly, because it takes too much cpu power

                          buffer[ ( buffer_pos + ( i / 2 )  ) ] = (p[ i + 2 ] + p[ i + 3 ] * 256);
                          buffer[ ( buffer_pos + ( i / 2 ) + 1  ) ] = (p[ i + 0 ] + p[ i + 1 ] * 256);
                        */

                        /* Left channel */
                        buffer[ buffer_pos + i ] = p[ i + 2 ];
                        buffer[ buffer_pos + i + 1 ] = p[ i + 3 ];

                        /* Right channel */
                        buffer[ buffer_pos + i + 2 ] = p[ i ];
                        buffer[ buffer_pos + i + 3 ] = p[ i + 1 ];

                }
                SDL_UnlockAudio();
                buffer_pos += i;
        }
#ifdef DEBUG
        else
        {
                fprintf(stderr, "(EE) Audio buffer overflow.\n");
        }
#endif

  
        /* Time that should be sleeped to keep game in sync */
        int wait_time = 0;
         
       /* And then syncronization */

       /* If buffer is running slow we speed up the game a bit. Actually we skip the syncronization. */
       if(buffer_pos < LOW_BUFFER_LOAD_LEVEL)
       {
               //wait_time -= (LOW_BUFFER_LOAD_LEVEL - buffer_pos);
               wait_time = -1;
       }
       if(wait_time != -1) {
               /* If for some reason game is runnin extremely fast and there is risk buffer is going to
                 overflow, we slow down the game a bit to keep sound smooth. The overspeed is caused
                 by inaccuracy in machines clock. */
               if(buffer_pos > HIGH_BUFFER_LOAD_LEVEL)
               {
                       wait_time += (float)(buffer_pos - HIGH_BUFFER_LOAD_LEVEL) / (float)(frequency / 250);
               }
       
               if(last_ticks + expected_ticks > SDL_GetTicks()) {
                       wait_time += (last_ticks + expected_ticks) - SDL_GetTicks();
#ifdef DEBUG
                       printf("wait_time: %i, Buffer: %i/%i\n", wait_time, buffer_pos, PRIMARY_BUFFER_SIZE);
#endif
                       SDL_Delay(wait_time);
               }
       
       }
       last_ticks = SDL_GetTicks();
       expected_ticks = ((float)(LenReg) / (float)(frequency / 250));
}

EXPORT DWORD CALL
AiReadLength( void )
{
   return 0;
}

EXPORT void CALL
CloseDLL( void )
{
}

EXPORT void CALL
DllAbout( HWND hParent )
{
#ifdef USE_GTK
   char tMsg[256];
   GtkWidget *dialog, *label, *okay_button;
   
   dialog = gtk_dialog_new();
   sprintf(tMsg,"Mupen64 SDL Sound Plugin v1.1 by JttL");
   label = gtk_label_new(tMsg);
   okay_button = gtk_button_new_with_label("OK");
   
   gtk_signal_connect_object(GTK_OBJECT(okay_button), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(dialog));
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
		     okay_button);
   
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
		     label);
   gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
   gtk_widget_show_all(dialog);
#else
   char tMsg[256];
   sprintf(tMsg,"Mupen64 SDL Sound Plugin 1.1 by JttL");
   fprintf(stderr, "About\n%s\n", tMsg);
#endif
}

EXPORT void CALL
DllConfig ( HWND hParent )
{
}

EXPORT void CALL
DllTest ( HWND hParent )
{
}

EXPORT void CALL
GetDllInfo( PLUGIN_INFO * PluginInfo )
{
   PluginInfo->Version = 0x0101;
   PluginInfo->Type    = PLUGIN_TYPE_AUDIO;
   sprintf(PluginInfo->Name,"JttL's SDL plugin");
   PluginInfo->NormalMemory  = TRUE;
   PluginInfo->MemoryBswaped = TRUE;
}

EXPORT BOOL CALL
InitiateAudio( AUDIO_INFO Audio_Info )
{
   AudioInfo = Audio_Info;
   return TRUE;
}
#ifdef DEBUG
static int underrun_count = 0;
#endif
void my_audio_callback(void *userdata, Uint8 *stream, int len)
{

        if(buffer_pos > len)
        {
                memcpy(stream, buffer, len);
                memmove(buffer, &buffer[ len ], buffer_pos  - len);

                buffer_pos = buffer_pos - len ;
        }
        else
        {          
#ifdef DEBUG
                 underrun_count++;
                 fprintf(stderr, "Audio buffer underrun (%i).\n",underrun_count);
#endif
                 memcpy(stream, buffer, buffer_pos );
                 buffer_pos = 0;
        }
        
}
EXPORT void CALL RomOpen()
{
        printf("(II) JttL's sound plugin version 1.1\n");
        printf("(II) Initializing SDL audio subsystem...\n");
        if(SDL_Init(SDL_INIT_AUDIO | SDL_INIT_TIMER) < 0)
        {
                fprintf(stderr, "(EE) Failed to initialize SDL audio subsystem.\n(EE) Forcing exit.\n");
                exit(-1);
        }
        else
        {
 
                atexit(SDL_Quit);
        }
        InitializeAudio( DEFAULT_FREQUENCY );

        if(buffer == NULL)
        {
                printf("(II) Allocating memory for audio buffer: %i bytes.\n", PRIMARY_BUFFER_SIZE*sizeof(Uint8));
                buffer = (Uint8*)malloc(PRIMARY_BUFFER_SIZE*sizeof(Uint8));
                SDL_PauseAudio(0);
        }

}               
void InitializeAudio(int freq)
{
        frequency = freq; //This is important for the sync
        if(hardware_spec != NULL) free(hardware_spec);
        SDL_PauseAudio(1);
        SDL_CloseAudio();

        /* Prototype of our callback function */
        void my_audio_callback(void *userdata, Uint8 *stream, int len);

        /* Open the audio device */
        SDL_AudioSpec *desired, *obtained;
        /* Allocate a desired SDL_AudioSpec */
        desired = malloc(sizeof(SDL_AudioSpec));
        /* Allocate space for the obtained SDL_AudioSpec */
        obtained = malloc(sizeof(SDL_AudioSpec));
        /* 22050Hz - FM Radio quality */
        desired->freq=freq;
        /* 16-bit signed audio */
        desired->format=AUDIO_U16SYS;
        /* Stereo */
        desired->channels=2;
        /* Large audio buffer reduces risk of dropouts but increases response time */
        desired->samples=SECONDARY_BUFFER_SIZE;

        /* Our callback function */
        desired->callback=my_audio_callback;
        desired->userdata=NULL;

        /* Open the audio device */
        if ( SDL_OpenAudio(desired, obtained) < 0 ){
                fprintf(stderr, "(EE) Couldn't open audio: %s\n", SDL_GetError());
                exit(-1);
        }
        /* desired spec is no longer needed */

        free(desired);
        hardware_spec=obtained;
        SDL_PauseAudio(0);

}
EXPORT void CALL
RomClosed( void )
{
   printf("(II) Cleaning up SDL sound plugin...\n");
   SDL_PauseAudio(1);
   if(buffer != NULL) free(buffer);
   if(hardware_spec != NULL) free(hardware_spec);
   hardware_spec = NULL;
   buffer = NULL;
   SDL_Quit();
}

EXPORT void CALL
ProcessAlist( void )
{
}
